import unittest
import clusto.util


class TestUtil(unittest.TestCase):
    def test_batch_against_iterable(self):
        i = clusto.util.batch(xrange(100), 10)
        first_batch = None
        last_batch = None
        num_batches = 0
        for batch_iter in i:
            num_batches += 1
            if first_batch is None:
                first_batch = list(batch_iter)
            last_batch = list(batch_iter)

        self.assertEqual(num_batches, 10)
        self.assertEqual(len(first_batch), 10)
        self.assertEqual(len(last_batch), 10)
        self.assertEqual(first_batch,
                         [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(last_batch,
                         [90, 91, 92, 93, 94, 95, 96, 97, 98, 99])

    def test_batch_unequal_lengths(self):
        i = clusto.util.batch(xrange(97), 10)
        first_batch = None
        last_batch = None
        num_batches = 0
        for batch_iter in i:
            num_batches += 1
            if first_batch is None:
                first_batch = list(batch_iter)
            last_batch = list(batch_iter)

        self.assertEqual(num_batches, 10)
        self.assertEqual(len(first_batch), 10)
        self.assertEqual(len(last_batch), 7)
        self.assertEqual(first_batch,
                         [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        self.assertEqual(last_batch,
                         [90, 91, 92, 93, 94, 95, 96])

    def test_batch_empty_list(self):
        i = clusto.util.batch([], 10)
        num_batches = 0
        for batch_iter in i:
            num_batches += 1
        self.assertEqual(num_batches, 0)

    def test_batch_zero_batch_length(self):
        i = clusto.util.batch(xrange(100), 0)
        with self.assertRaises(AssertionError):
            for g in i:
                pass



#!/usr/bin/env python
try:
    import simplejson as json
except ImportError:
    import json

from webob import Request, Response
from traceback import format_exc
from urllib import unquote_plus
import new
import re
import os

from clusto.drivers import Driver, IPManager
import clusto.script_helper
import clusto

from clusto.services.config import conf, get_logger
log = get_logger('clusto.http', 'INFO')

conf = clusto.script_helper.load_config(os.environ.get('CLUSTOCONFIG', '/etc/clusto/clusto.conf'))
clusto.connect(conf)


def unclusto(obj, prefetch_attrs=None):
    '''
    Convert an object to a representation that can be safely serialized into
    JSON.
    '''
    if type(obj) in (str, unicode, int) or obj is None:
        return obj
    if type(obj) in (list, dict):
        return json.dumps(obj)
    if isinstance(obj, clusto.Attribute):
        return {
            'key': obj.key,
            'value': unclusto(obj.value),
            'subkey': obj.subkey,
            'number': obj.number,
            'datatype': obj.datatype
        }
    if issubclass(obj.__class__, Driver):
        if not prefetch_attrs:
            return '/%s/%s' % (obj.type, obj.name)

        result = {
            'attrs': []
        }
        for prefetch_attr in prefetch_attrs:
            for attr in obj.attrs(**prefetch_attr):
                result['attrs'].append(unclusto(attr))
        return {'/%s/%s' % (obj.type, obj.name): result}
    return str(obj)


def dumps(request, obj, **kwargs):
    result = json.dumps(obj, indent=2, sort_keys=True, **kwargs)
    if 'callback' in request.params:
        callback = request.params['callback']
        result = '%s(%s)' % (callback, result)
        content_type = 'text/javascript'
    else:
        content_type = 'application/json'
    return Response(result, content_type=content_type, **kwargs)


def loads(request, obj):
    return json.loads(obj)


class EntityAPI(object):
    def __init__(self, obj):
        self.obj = obj
        self.url = '/%s/%s' % (self.obj.type, self.obj.name)

    def _dict(self):
        result = {
            'object': self.url,
            'driver': self.obj.driver,
        }

        attrs = []
        for x in self.obj.attrs():
            attrs.append(unclusto(x))
        result['attrs'] = attrs
        result['contents'] = [unclusto(x) for x in self.obj.contents()]
        result['parents'] = [unclusto(x) for x in self.obj.parents()]
        result['actions'] = [x for x in dir(self) if not x.startswith('_') and callable(getattr(self, x))]
        return result

    def addattr(self, request):
        '''
        Add an attribute to this object.

        Requires HTTP parameters "key" and "value"
        Optional parameters are "subkey", "number" and "datatype"
        where "datatype" could be "relation", "int" or string by default
        '''
        kwargs = dict(request.params.items())
        datatype = kwargs.get('datatype', None)
        kwargs.pop('datatype', None)

        if datatype == 'relation':
            try:
                kwargs['value'] = clusto.get_by_name(kwargs['value'].rsplit('/', 1)[1])
            except LookupError:
                return Response(status=400, body='400 Element Not Found\n')
        elif datatype == 'int':
            kwargs['value'] = int(kwargs['value'])
        else:
            kwargs['value'] = str(kwargs['value'])

        self.obj.add_attr(**kwargs)
        return self.show(request)

    def setattr(self, request):
        '''
        Set an attribute on this object.

        Requires HTTP parameters "key" and "value"
        Optional parameters are "subkey", "number" and "datatype"
        where "datatype" could be "relation", "int" or string by default
        '''
        kwargs = dict(request.params.items())
        datatype = kwargs.get('datatype', None)
        kwargs.pop('datatype', None)

        if datatype == 'relation':
            try:
                kwargs['value'] = clusto.get_by_name(kwargs['value'].rsplit('/', 1)[1])
            except LookupError:
                return Response(status=400, body='400 Element Not Found\n')
        elif datatype == 'int':
            kwargs['value'] = int(kwargs['value'])
        else:
            kwargs['value'] = str(kwargs['value'])

        if 'number' in kwargs:
            kwargs['number'] = int(kwargs['number'])
        self.obj.set_attr(**kwargs)
        return self.show(request)

    def delattr(self, request):
        '''
        Delete an attribute from this object.

        Requires HTTP parameter "key"
        '''
        self.obj.del_attrs(request.params['key'])
        return self.show(request)

    def attrs(self, request):
        '''
        Query attributes from this object.
        '''
        result = {
            'attrs': []
        }
        kwargs = dict(request.params.items())
        for attr in self.obj.attrs(**kwargs):
            result['attrs'].append(unclusto(attr))
        return dumps(request, result)

    def insert(self, request):
        '''
        Insert an object into this object

        Requires a "device" attribute, which is the absolute URL path to the
        object to be inserted.

        For example: /pool/examplepool/insert?object=/server/exampleserver
        '''
        device = request.params['object'].strip('/').split('/')[1]
        device = clusto.get_by_name(device)
        self.obj.insert(device)
        return self.show(request)

    def remove(self, request):
        '''
        Remove a device from this object.

        Requires a "device" attribute, which is the absolute URL path to the
        object to be removed.

        For example: /pool/examplepool/remove?object=/server/exampleserver
        '''
        device = request.params['object'].strip('/').split('/')[1]
        device = clusto.get_by_name(device)
        self.obj.remove(device)
        return self.show(request)

    def show(self, request, prefetch_attrs=None):
        '''
        Returns attributes and actions available for this object.
        '''
        result = {}
        result['object'] = self.url
        result['driver'] = self.obj.driver

        attrs = []
        for x in self.obj.attrs():
            attrs.append(unclusto(x))
        result['attrs'] = attrs
        result['contents'] = [unclusto(x, prefetch_attrs) for x in self.obj.contents()]
        result['parents'] = [unclusto(x, prefetch_attrs) for x in self.obj.parents()]
        result['actions'] = [x for x in dir(self) if not x.startswith('_') and callable(getattr(self, x))]

        return dumps(request, result)


class PortInfoAPI(EntityAPI):
    def ports(self, request):
        '''
        Returns a list of ports within this object and any information about
        connections between those ports and other devices.
        '''
        result = {}
        result['object'] = self.url
        result['ports'] = []
        for port in self.obj.port_info_tuples:
            porttype, portnum, otherobj, othernum = [unclusto(x) for x in port]
            result['ports'].append({
                'type': porttype,
                'num': portnum,
                'other': otherobj,
                'othernum': othernum,
            })

        return dumps(request, result)

    def set_port_attr(self, request):
        kwargs = {}
        params = request.params
        for arg in ('porttype', 'portnum', 'key', 'value'):
            if not arg in params:
                return Response(status=400, body='Required argument "%s" is missing\n' % arg)
            kwargs[arg] = params[arg]

        try:
            kwargs['portnum'] = int(kwargs['portnum'])
        except ValueError:
            return Response(status=400, body='portnum must be an integer\n')

        self.obj.set_port_attr(**kwargs)
        return Response(status=200)

    def get_port_attr(self, request):
        kwargs = {}
        params = request.params
        for arg in ('porttype', 'portnum', 'key'):
            if not arg in params:
                return Response(status=400, body='Required argument "%s" is missing\n' % arg)
            kwargs[str(arg)] = params[arg]

        try:
            kwargs['portnum'] = int(kwargs['portnum'])
        except ValueError:
            return Response(status=400, body='portnum must be an integer\n')

        return dumps(request, self.obj.get_port_attr(**kwargs))


class RackAPI(EntityAPI):
    def insert(self, request):
        '''
        Insert a device into a rack.

        This method is overridden to require a "ru" parameter in addition to
        "device"

        Example: /rack/examplerack/insert?object=/server/exampleserver&ru=6
        '''
        device = request.params['object'].strip('/').split('/')[1]
        device = clusto.get_by_name(device)
        self.obj.insert(device, int(request.params['ru']))
        return self.contents(request)


class ResourceAPI(EntityAPI):
    def allocate(self, request):
        '''
        Allocate a new object of the given type
        '''
        driver = clusto.DRIVERLIST[request.params['driver']]
        device = self.obj.allocate(driver)
        return dumps(request, unclusto(device), status=201)


class QueryAPI(object):
    @classmethod
    def get_entities(self, request):
        kwargs = {'attrs': []}
        for k, v in request.params.items():
            if k in ('format', 'callback'): continue
            v = loads(request, unquote_plus(v))
            kwargs[str(k)] = v

        attrs = []
        for attr in kwargs['attrs']:
            attrs.append(dict([(str(k), v) for k, v in attr.items()]))
        kwargs['attrs'] = attrs

        result = [unclusto(x) for x in clusto.get_entities(**kwargs)]
        return dumps(request, result)

    @classmethod
    def get_by_name(self, request):
        if not 'name' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specify a "name" parameter\n')
        name = request.params['name']
        obj = clusto.get_by_name(name)
        api = EntityAPI(obj)
        prefetch_attrs = None
        if 'prefetch_attrs' in request.params:
            # Prefetch all the matching attributes while pulling data
            # Prefetch format would like something like
            #   prefetch_attrs=[{'key': 'disk', 'subkey': 'make'}, {'key': 'system', 'subkey': 'version'}]
            prefetch_attrs = loads(request, request.params['prefetch_attrs'])
        return api.show(request, prefetch_attrs)

    @classmethod
    def get(self, request):
        if not 'name' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specifiy a "name" parameter\n')
        name = request.params['name']

        result = clusto.get(name)
        if result is None:
            return Response(status=404, body='404 Not Found\n')

        result = [EntityAPI(x)._dict() for x in result]
        return dumps(request, result)

    @classmethod
    def get_from_pools(self, request):
        if not 'pools' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specify at least one pool\n')
        pools = request.params['pools'].split(',')

        if 'types' in request.params:
            clusto_types = request.params['types'].split(',')
        else:
            clusto_types = None

        prefetch_attrs = None
        if 'prefetch_attrs' in request.params:
            # Prefetch all the matching attributes while pulling data
            # Prefetch format would like something like
            #   prefetch_attrs=[{'key': 'disk', 'subkey': 'make'}, {'key': 'system', 'subkey': 'version'}]
            prefetch_attrs = loads(request, request.params['prefetch_attrs'])

        result = [unclusto(x, prefetch_attrs) for x in clusto.get_from_pools(pools, clusto_types)]
        return dumps(request, result)

    @classmethod
    def get_ip_manager(self, request):
        if not 'ip' in request.params:
            return Response(status=400, body='400 Bad Request\nYou must specify an IP\n')

        try:
            ipman = IPManager.get_ip_manager(request.params['ip'])
        except:
            return Response(status=404, body='404 Not Found\n')
        return dumps(request, unclusto(ipman))


class ClustoApp(object):
    def __init__(self):
        self.urls = [
            ('^/favicon.ico$',
                self.notfound),
            ('^/search$',
                self.search),
            ('^/query/(?P<querytype>[a-z_]+)',
                self.query_delegate),
            ('^/(?P<objtype>\w+)/(?P<name>[-\w0-9]+)/(?P<action>\w+)',
                self.action_delegate),
            ('^/(?P<objtype>\w+)/(?P<name>[-\w0-9]+)',
                self.action_delegate),
            ('^/(?P<objtype>\w+)',
                self.types_delegate),
            ('^/',
                self.default_delegate),
        ]
        self.urls = [(re.compile(pattern), obj) for pattern, obj in self.urls]

        self.types = {
            'server': PortInfoAPI,
            'consoleserver': PortInfoAPI,
            'networkswitch': PortInfoAPI,
            'powerstrip': PortInfoAPI,
            'rack': RackAPI,
            'resource': ResourceAPI,
        }

    def default_delegate(self, request, match):
        types = ['/' + x for x in clusto.typelist.keys()]
        types.sort()
        return dumps(request, types)

    def types_delegate(self, request, match):
        objtype = match.groupdict()['objtype']
        result = []
        for obj in clusto.get_entities(clusto_types=(objtype,)):
            result.append(unclusto(obj))
        return dumps(request, result)

    def action_delegate(self, request, match):
        if request.method == 'GET':
            return self.get_action(request, match)

        if request.method == 'POST':
            return self.post_action(request, match)

        if request.method == 'DELETE':
            return self.delete_action(request, match)

        return Response(status=501, body='501 Not Implemented\n')

    def query_delegate(self, request, match):
        querytype = match.groupdict()['querytype']

        if hasattr(QueryAPI, querytype):
            method = getattr(QueryAPI, querytype)
            return method(request)
        else:
            return Response(status=400, body='400 Bad Request\nThere is no such query\n')

    def post_action(self, request, match):
        name = request.path_info.strip('/')
        if name.count('/') != 1:
            return Response(status=400, body='400 Bad Request\nYou may only create objects, not types or actions\n')
        objtype, objname = name.split('/', 1)

        try:
            obj = clusto.get_by_name(objname)
            if obj:
                return Response(status=409, body='409 Conflict\nObject already exists\n')
        except LookupError: pass

        obj = clusto.driverlist[objtype](objname)

        obj = EntityAPI(obj)
        response = obj.show(request)
        response.status = 201
        return response

    def delete_action(self, request, match):
        name = request.path_info.strip('/')
        if name.count('/') != 1:
            return Response(status=400, body='400 Bad Request\nYou may only delete objects, not types or actions\n')
        objtype, objname = name.split('/', 1)

        try:
            obj = clusto.get_by_name(objname)
        except LookupError:
            return Response(status=404, body='404 Not Found\n')

        clusto.delete_entity(obj.entity)
        return Response(status=200, body='200 OK\nObject deleted\n')

    def get_action(self, request, match):
        group = match.groupdict()
        try:
            obj = clusto.get_by_name(group['name'])
        except LookupError:
            return Response(status=404, body='404 Not Found\n')

        if obj.type != group['objtype']:
            response = Response(status=302)
            response.headers.add('Location', str('/%s/%s/%s' % (obj.type, obj.name, group.get('action', ''))))
            return response

        action = group.get('action', 'show')
        handler = self.types.get(group['objtype'], EntityAPI)
        if not obj:
            obj = clusto.get_by_name(group['name'])
        h = handler(obj)
        if hasattr(h, action):
            f = getattr(h, action)
            response = f(request)
        else:
            response = Response(status=404, body='404 Not Found\nInvalid action\n')
        return response

    def search(self, request, match):
        query = request.params.get('q', None)
        if not query:
            return Response(status=400, body='400 Bad Request\nNo query specified\n')

        result = []
        for obj in clusto.get_entities():
            if obj.name.find(query) != -1:
                result.append(unclusto(obj))
        return dumps(request, result)

    def notfound(self, request, match):
        return Response(status=404)

    def __call__(self, environ, start_response):
        request = Request(environ)
        response = Response(status=404, body='404 Not Found\nUnmatched URL\n')

        for pattern, handler in self.urls:
            match = pattern.match(request.path_info)
            if match:
                try:
                    response = handler(request, match)
                except:
                    response = Response(status=500, body=format_exc(),
                                        content_type='text/plain')
                break
        return response(environ, start_response)


application = ClustoApp()


if __name__ == '__main__':
    from wsgiref.simple_server import make_server
    server = make_server('0.0.0.0', 9996, application)
    log.info('clusto httpd loaded using wsgiref (debug mode)')
    server.serve_forever()



#!/usr/bin/env python
"""Display clusto objects, recursively, with attributes, and/or in color."""

import sys
from itertools import chain

import clusto
from clusto import script_helper


class Colors(object):

    """ANSI color escape sequences class."""

    mapping = {'BRIGHT': '\033[1m',
               'GREEN': '\033[32m',
               'MAGENTA': '\033[35m',
               'YELLOW': '\033[33m',
               'RESET': '\033[0m'}

    def __init__(self, enabled=False):
        """Construct instance with color enabled or disabled."""
        # Could default to true if stdout is a tty and not on Windows,
        # but for now we don't try to guess.
        self.enabled = enabled

    def __getattr__(self, name):
        """Return ANSI color escape sequence if color is enabled."""
        if self.enabled:
            return self.mapping[name]
        else:
            assert name in self.mapping
            return ''


class Tree(script_helper.Script):

    """
    Display clusto objects, recursively, with attributes, and/or in color.

    This script queries for a specified clusto object and attributes and
    optionally also queries parent or contained objects, recursively.
    It displays a simple, human-readable indented tree, with different
    colors for clusto keys, subkeys, and values if color is enabled.
    """

    def print_obj(self, obj, attrs, indent=0, color=False):
        """Print indented object, optionally with attributes or in color."""
        colors = Colors(enabled=color)
        indent_txt = "\t" * indent

        # Clusto object name.
        txt = "{indent}* {name}".format(
            indent=indent_txt,
            name=colors.BRIGHT + obj.name + colors.RESET)

        # Clusto object attributes.
        if 'ALL' in attrs:
            attributes = obj.attrs()
        else:
            attributes = chain.from_iterable(obj.attrs(attr)
                                             for attr in attrs)
        for x in attributes:
            if x.subkey:
                subkey_txt = ".{subkey}".format(
                    subkey=colors.MAGENTA + str(x.subkey) + colors.RESET
                )
            else:
                subkey_txt = ""
            txt += "\n{indent}| {key}{subkey} = {value}".format(
                indent=indent_txt,
                key=colors.GREEN + str(x.key) + colors.RESET,
                subkey=subkey_txt,
                value=colors.YELLOW + str(x.value) + colors.RESET
            )

        print(txt)

    def print_tree(self, root, attrs, direction, indent=0, color=False):
        """Print parent or contained object attributes, recursively."""
        # Alternatively, we could call obj.attrs(merge_container_attrs=True),
        # and use the entity attribute of each clusto attribute to
        # recreate the tree.
        for parent in getattr(root, direction)():
            self.print_obj(parent, attrs, indent=indent, color=color)
            self.print_tree(parent, attrs, direction, indent=indent + 1,
                            color=color)

    def run(self, args):
        """Execute script, passing script arguments to methods."""
        obj = clusto.get_by_name(args.obj)
        attrs = args.attrs
        self.print_obj(obj, attrs, indent=0, color=args.color)
        if args.parents or args.contents:
            if args.parents:
                direction = 'parents'
            else:
                direction = 'contents'
            self.print_tree(obj, attrs, direction, indent=1,
                            color=args.color)

    def _add_arguments(self, parser):
        parser.add_argument('obj', metavar='object', help="clusto object")
        parser.add_argument('attrs', metavar='attributes', nargs='*',
                            help="clusto attribute keys, ALL for all")
        direction = parser.add_mutually_exclusive_group()
        direction.add_argument('-p', '--parents', action='store_true',
                               default=False,
                               help="query parent objects, recursively")
        direction.add_argument('-c', '--contents', action='store_true',
                               default=False,
                               help="query contained objects, recursively")
        parser.add_argument('-C', '--color', action='store_true',
                            default=False, help="enable color output")


def main():
    """Execute script with clusto script_helper."""
    attr, args = script_helper.init_arguments(Tree)
    return attr.run(args)

if __name__ == '__main__':
    sys.exit(main())



"""Common utility functions for clusto"""
import itertools


def batch(iterable, size):
    """Returns an iterable of iterables that each yield size items."""
    assert size != 0, 'size must be nonzero'
    c = itertools.count()
    for _, g in itertools.groupby(iterable, lambda x: c.next()/size):
        yield g



#!/usr/bin/env python
# -*- mode: python; sh-basic-offset: 4; indent-tabs-mode: nil; coding: utf-8 -*-
# vim: tabstop=4 softtabstop=4 expandtab shiftwidth=4 fileencoding=utf-8

import argparse
import sys

import clusto
from clusto import drivers
from clusto import script_helper

class Reboot(script_helper.Script):
    '''
    This will reboot a given server or IP address
    '''

    def __init__(self):
        script_helper.Script.__init__(self)

    def _add_arguments(self, parser):
        parser.add_argument('--batch', action='store_true', default=False,
            help='Batch mode, don\'t prompt for confirmation')
        parser.add_argument('--method', default=None,
            help='Use the given method to reboot this server (eg. ipmi)')
        parser.add_argument('server', nargs='+',
            help='Server name or IP address')

    def add_subparser(self, subparsers):
        parser = self._setup_subparser(subparsers)
        self._add_arguments(parser)

    def confirm(self, server):
        print 'Parents: %s' % ' '.join([x.name for x in server.parents()])
        response = raw_input('Are you sure you want to reboot %s (yes/no)? ' % server.name)
        if response == 'yes':
            return True
        else:
            return False

    def run(self, args):
        for name in args.server:
            server = clusto.get(name)
            if not server:
                self.error('%s does not exist' % name)
                continue

            server = server[0]

            if not hasattr(server, 'reboot'):
                self.error('%s does not implement reboot()' % server.name)
                return -1

            if not args.batch:
                if not self.confirm(server):
                    return -1

            kwargs = {}
            if args.method is not None:
                kwargs['method'] = args.method
            server.reboot(**kwargs)


def main():
    reboot = Reboot()
    parent_parser = script_helper.setup_base_parser()
    this_parser = argparse.ArgumentParser(parents=[parent_parser],
        description=reboot._get_description())
    reboot._add_arguments(this_parser)
    args = this_parser.parse_args()
    reboot.init_script(args=args, logger=script_helper.get_logger(args.loglevel))
    return(reboot.run(args))


if __name__ == '__main__':
    sys.exit(main())

